// Decorator.cpp : Implementation of CDecorator
#include "stdafx.h"
#include "Decorator.h"



#define VERIFY_INIT   { if (!m_isInitialized) return E_DECORATOR_UNINITIALIZED; }
#define VERIFY_LOCSET { if (!m_isLocSet) return E_DECORATOR_LOCISNOTSET; }

CDecoratorUtil	d_util;

#define SAMPLE_SIZEX	400
#define SAMPLE_SIZEY	200

/////////////////////////////////////////////////////////////////////////////
// CDecorator
STDMETHODIMP CDecorator::Initialize(IMgaProject *project, IMgaMetaPart *metaPart, IMgaFCO *obj)
{

	AFX_MANAGE_STATE(AfxGetStaticModuleState( ));//magic words to make CMaskedBitmap to work
	//
	// TODO: read all important data from MGA and cache them for later use
	//
	m_mgaFco = obj;		// obj == NULL, if we are in the PartBrowser
	
	if (!GetMetaFCO(metaPart, m_metaFco)) 
	{
		return E_DECORATOR_INIT_WITH_NULL;
	}
	else 
	{
		m_isInitialized = true;
	}

	if (SetupImported())
	{
		m_isImported = true;
	}

	if (m_mgaFco) 
	{ //if there's an FCO (this obj not in part browser)
		objtype_enum	objtype;
		COMTHROW(m_mgaFco->get_ObjType(&objtype));
		if (objtype == OBJTYPE_REFERENCE) 
		{
			if(m_isImported)
				m_copyBitmap.ReadFromResource(IDB_BITMAP_IMPORT);
			else
				m_copyBitmap.ReadFromResource(IDB_BITMAP_COPY);
			m_isCopy = true;
			CComPtr<IMgaFCO> mgaFco = m_mgaFco;
			while(objtype == OBJTYPE_REFERENCE) 
			{//recursively find object referred to
				CComPtr<IMgaReference> ref;
				COMTHROW(mgaFco.QueryInterface(&ref));
				mgaFco = NULL;
				COMTHROW(ref->get_Referred(&mgaFco));
				if (mgaFco) 
				{
					COMTHROW(mgaFco->get_ObjType(&objtype));
				}
				else 
				{
					objtype = OBJTYPE_NULL; //referred chain ends in null
				}
			}
			if (objtype == OBJTYPE_NULL) 
			{//no object referred to
				
			}
			else 
			{//there is an object referred to, set the meta to its meta, so the Setup Functions will
				  // correctly prepare it for drawing, m_isCopy will make sure a referring arrow is drawn
				m_mgaFco = mgaFco;
				CComPtr<IMgaMetaFCO> metaFco;
				metaFco = NULL;
				COMTHROW(m_mgaFco->get_Meta(&metaFco));
				if(metaFco)
				{
					m_metaFco = metaFco;
				}

			}
		}
		else 
		{
			m_isCopy = false;
		}
		CComBSTR bstr;
		COMTHROW(m_mgaFco->get_Name(&bstr));
		m_name = bstr;

		//check if this object is abstract
		bool isAbstract = false;
		if (GetAttribute(isAbstract, MOF_ABSTRACT_ATTR, m_mgaFco) && isAbstract) 
			m_isAbstract = true;

	}
	else 
	{ // no m_mgaFco
		objtype_enum objtype;
		COMTHROW(m_metaFco->get_ObjType(&objtype));
		if (objtype == OBJTYPE_REFERENCE) 
		{
			m_copyBitmap.ReadFromResource(IDB_BITMAP_COPY);
			m_isCopy = true;
		}
		else 
		{
			m_isCopy = false;
		}
		CComBSTR bstr;
		COMTHROW(m_metaFco->get_DisplayedName(&bstr));
		if (bstr.Length() == 0 ) 
		{
			bstr.Empty();
			COMTHROW(m_metaFco->get_Name(&bstr));
		}
		m_name = bstr;
	}

	if (m_isImported && !m_mgaFco)
		return S_OK;
	
	if (SetupInheritance()) {
		if(!GetColorPreference(m_color,COLOR_PREF)) {
			m_color = GME_BLACK_COLOR;
		}
		return S_OK;
	}

	if (SetupAssociation()) {
		if(!GetColorPreference(m_color,COLOR_PREF)) {
			m_color = GME_BLACK_COLOR;
		}
		return S_OK;
	}

	if (SetupConstraint()) {
		
		return S_OK;
	}

	if (SetupType()) {

		CalcRelPositions();
		return S_OK;
	}

	if (SetupAttrParamSField())
	{
		CalcRelPositions();
		return S_OK;
	}

	if (SetupOperation())
	{
		CalcRelPositions();
		return S_OK;
	}

	if (SetupTag()) 
	{
		return S_OK;
	}

	if (SetupPackage())
	{
		CalcRelPositions();
		return S_OK;
	}

	if (SetupConst())
	{
		CalcRelPositions();
		return S_OK;
	}

	if (SetupClass())
	{
		CalcRelPositions();
		return S_OK;
	}

	if (SetupException())
	{
		CalcRelPositions();
		return S_OK;
	}

	if (SetupAspect())
	{
		if(!m_isSameAspect)
			CalcRelPositions();
		return S_OK;
	}
	
	if(!GetColorPreference(m_color,COLOR_PREF)) {
		m_color = GME_BLACK_COLOR;
	}
	if(!GetColorPreference(m_nameColor,NAME_COLOR_PREF)) {
		m_nameColor = GME_BLACK_COLOR;
	}

	CalcRelPositions();

	return S_OK;
}

STDMETHODIMP CDecorator::Destroy()
{
	//
	// TODO: At least free all references to MGA objects
	//
	VERIFY_INIT;
	m_isInitialized = false;
	m_isLocSet = false;
	m_metaFco = NULL; m_mgaFco = NULL;
	return S_OK;
}

STDMETHODIMP CDecorator::GetMnemonic(BSTR *mnemonic)
{	
	//
	// TODO: Return the logical name of the decorator (currently not used by GME)
	//
	*mnemonic = CComBSTR(DECORATOR_NAME).Detach();
	return S_OK;
}

STDMETHODIMP CDecorator::GetFeatures(feature_code *features)
{	
	//
	// TODO: Return supported features (combine multiple features with bitwise-OR)
	// Available feature codes are found in MgaDecorator.idl
	// (curently not used by GME)
	*features = 0;
	return S_OK;
}

STDMETHODIMP CDecorator::SetParam(BSTR name, VARIANT value)
{
	//
	// TODO:  Parse and set all supported parameters, otherwise return error
	// (currently all values are BSTR type)
	//
	VERIFY_INIT;
	return E_DECORATOR_UNKNOWN_PARAMETER;
}

STDMETHODIMP CDecorator::GetParam(BSTR name, VARIANT* value)
{
	//
	// TODO: Return values of supported and previously set parameters, otherwise return error
	// (currently GME does not use this method)
	//
	VERIFY_INIT;
	return E_DECORATOR_UNKNOWN_PARAMETER;
}

STDMETHODIMP CDecorator::SetActive(VARIANT_BOOL isActive)
{
	//
	// TODO: If isActive==VARIANT_FALSE, draw your object in GME_GREYED_OUT, otherwise use the color of the object
	//
	VERIFY_INIT;
	m_isActive = (isActive != VARIANT_FALSE);
	return S_OK;
}

STDMETHODIMP CDecorator::GetPreferredSize(long* sizex, long* sizey) 
{
	//
	// TODO: Give GME a hint about the object size. Do not expect GME to take it into account
	//
	VERIFY_INIT;
	*sizex = m_calcSize.cx;
	*sizey = m_calcSize.cy;
	return S_OK;
}


STDMETHODIMP CDecorator::SetLocation(long sx, long sy, long ex, long ey)
{
	//
	// TODO: Draw the object exactly to the this location later
	//
	VERIFY_INIT;
	m_sx = sx;
	m_sy = sy;
	m_ex = ex;
	m_ey = ey;
	m_isLocSet = true;
	return S_OK;
}

STDMETHODIMP CDecorator::GetLocation(long *sx, long *sy, long *ex, long *ey)
{
	//
	// TODO: Return previously set location parameters
	// (currently GME does not call this)
	// 
	VERIFY_INIT;
	VERIFY_LOCSET;
	*sx = m_sx;
	*sy = m_sy;
	*ex = m_ex;
	*ey = m_ey;
	return S_OK;
}

STDMETHODIMP CDecorator::GetLabelLocation(long *sx, long *sy, long *ex, long *ey)
{
	//
	// TODO: Return the location of the text box of your label if you support labels.
	// (currently GME does not call this)
	//
	VERIFY_INIT;
	VERIFY_LOCSET;
	return S_OK;
}

STDMETHODIMP CDecorator::GetPortLocation(IMgaFCO *fco, long *sx, long *sy, long *ex, long *ey)
{
	//
	// TODO: Return the location of the specified port if ports are supported in the decorator
	//
	VERIFY_INIT;
	VERIFY_LOCSET;

	return E_DECORATOR_PORTNOTFOUND;
}

STDMETHODIMP CDecorator::GetPorts(IMgaFCOs **portFCOs)
{
	//
	// TODO: Return a collection of mga objects represented as ports.
	//
	VERIFY_INIT;
	CComPtr<IMgaFCOs> coll;
	COMTHROW(coll.CoCreateInstance(OLESTR("Mga.MgaFCOs")));
	*portFCOs = coll.Detach();
	return S_OK;
}


STDMETHODIMP CDecorator::Draw(HDC hdc)
{
	//
	// TODO: Draw your object. Saving & restoring the HDC is a good practice, however it is not a strict requirement
	// Use d_util for brushes/pens/fonts.
	//
	VERIFY_INIT;
	VERIFY_LOCSET;

	CDC dc;
	dc.Attach(hdc);
	CRect rect(m_sx, m_sy, m_ex, m_ey);
	//d_util.DrawFlatBox(&dc, rect, m_color, m_color);

	double scalex = ((double)(m_ex - m_sx)) / (m_calcSize.cx);
	double scaley = ((double)(m_ey - m_sy)) / (m_calcSize.cy);
	

	if (m_isInheritance) { //the inheritance triangle
		dc.SelectObject(d_util.GetPen(m_isActive ? m_color : GME_GRAYED_OUT_COLOR));
		dc.SelectStockObject(NULL_BRUSH);
		dc.BeginPath();
		dc.MoveTo(m_sx, m_ey);
		dc.LineTo(m_ex, m_ey);
		dc.LineTo(m_sx + (m_ex-m_sx)/2, m_sy);
		dc.LineTo(m_sx, m_ey);
		dc.EndPath();
		dc.StrokeAndFillPath();
		if (!m_mgaFco) { //only draw the name in the partbrowser
			CPoint namePos(m_sx + ((m_ex - m_sx) / 2), m_ey);
			d_util.DrawText(&dc, m_name, namePos, d_util.GetFont(GME_PORTNAME_FONT), (m_isActive ? m_color : GME_GRAYED_OUT_COLOR), TA_TOP | TA_CENTER);
		}
		dc.Detach();
		return S_OK;
	}

	if (m_isSameAspect)
	{
		dc.SelectObject(d_util.GetPen(m_isActive ? m_color : GME_GRAYED_OUT_COLOR));
		dc.SelectStockObject(NULL_BRUSH);
		dc.BeginPath();
		int midx = (m_sx+m_ex)/2;
		int midy = (m_sy+m_ey)/2;
		dc.MoveTo(midx,m_sy);
		dc.LineTo(m_sx,midy);
		dc.LineTo(midx,m_ey);
		dc.LineTo(m_ex,midy);
		dc.LineTo(midx,m_sy);
		dc.EndPath();
		dc.StrokeAndFillPath();
		if (!m_mgaFco) { //only draw the name in the partbrowser
			CPoint namePos(m_sx + ((m_ex - m_sx) / 2), m_ey);
			d_util.DrawText(&dc, m_name, namePos, d_util.GetFont(GME_PORTNAME_FONT), (m_isActive ? m_color : GME_GRAYED_OUT_COLOR), TA_TOP | TA_CENTER);
		}
		dc.Detach();
		return S_OK;	
	}


	if (m_isAssociation) {//little circle for association
		dc.SelectObject(d_util.GetPen(m_isActive ? m_color : GME_GRAYED_OUT_COLOR));
		dc.SelectObject(d_util.GetBrush(m_isActive ? m_color : GME_GRAYED_OUT_COLOR));
		dc.Ellipse(m_sx, m_sy, m_ex, m_ey);
		
		//unlike the UML and MetaGME connector, association names get displayed anyways
		CPoint namePos(m_sx + ((m_ex - m_sx) / 2), m_ey);
		d_util.DrawText(&dc, m_name, namePos, d_util.GetFont(GME_PORTNAME_FONT), (m_isActive ? m_color : GME_GRAYED_OUT_COLOR), TA_TOP | TA_CENTER);
		dc.Detach();
		return S_OK;
	}

	if (m_isConstraint && m_constraintBitmap.IsValid()) {
		m_constraintBitmap.DrawTransparent(&dc, m_sx, m_sy, MOF_TRANSPARENT_COLOR, !m_isActive, GME_GRAYED_OUT_COLOR);
		CPoint namePos(m_sx + ((m_ex - m_sx) / 2), m_ey);
		d_util.DrawText(&dc, m_name, namePos, d_util.GetFont(GME_PORTNAME_FONT), (m_isActive ? m_color : GME_GRAYED_OUT_COLOR), TA_TOP | TA_CENTER);
		dc.Detach();
		return S_OK;
	}

	if (m_isTag && m_tagBitmap.IsValid()) { //same as constraint, different bitmap
		m_tagBitmap.DrawTransparent(&dc, m_sx, m_sy, MOF_TRANSPARENT_COLOR, !m_isActive, GME_GRAYED_OUT_COLOR);
		CPoint namePos(m_sx + ((m_ex - m_sx) / 2), m_ey);
		d_util.DrawText(&dc, m_name, namePos, d_util.GetFont(GME_PORTNAME_FONT), (m_isActive ? m_color : GME_GRAYED_OUT_COLOR), TA_TOP | TA_CENTER);
		dc.Detach();
		return S_OK;
	}

	if (m_isImported && m_importedBitmap.IsValid() && !m_mgaFco) { //same as constraint, different bitmap
		m_importedBitmap.DrawTransparent(&dc, m_sx, m_sy, MOF_TRANSPARENT_COLOR, !m_isActive, GME_GRAYED_OUT_COLOR);
		CPoint namePos(m_sx + ((m_ex - m_sx) / 2), m_ey);
		d_util.DrawText(&dc, m_name, namePos, d_util.GetFont(GME_PORTNAME_FONT), (m_isActive ? m_color : GME_GRAYED_OUT_COLOR), TA_TOP | TA_CENTER);
		dc.Detach();
		return S_OK;
	}

	if (m_isException) { 
		
		dc.SelectObject(d_util.GetPen(m_isActive ? m_color : GME_GRAYED_OUT_COLOR));
		dc.SelectStockObject(NULL_BRUSH);
		dc.Rectangle(m_sx, m_sy, m_ex, m_ey);

		dc.SelectObject(d_util.GetBrush(m_isActive ? m_color : GME_GRAYED_OUT_COLOR));
		dc.Ellipse((m_sx+m_ex)/2-MOF_EXCEPTION_RADIUS, m_sy+MOF_EXCEPTION_HEIGHT-2*MOF_EXCEPTION_RADIUS,
			(m_sx+m_ex)/2+MOF_EXCEPTION_RADIUS, m_sy+MOF_EXCEPTION_HEIGHT);
		dc.Ellipse((m_sx+m_ex)/2-MOF_EXCEPTION_RADIUS, m_sy+MOF_DECORATOR_MARGINY, (m_sx+m_ex)/2+MOF_EXCEPTION_RADIUS,
			m_sy+MOF_EXCEPTION_HEIGHT-2-2*MOF_EXCEPTION_RADIUS);
		
		CPoint namePos(m_sx + ((m_ex - m_sx) / 2), m_ey-MOF_DECORATOR_MARGINY);
		d_util.DrawText(&dc, m_name, namePos, d_util.GetFont(GME_PORTNAME_FONT), (m_isActive ? m_color : GME_GRAYED_OUT_COLOR), TA_BOTTOM | TA_CENTER);
		//dc.Detach();
		//return S_OK; //let fall thru to copyBitmap
	}

	if(m_isType || m_isStructField)
	{
		d_util.DrawFlatBox(&dc, rect, m_color, MOF_BLUE_COLOR);
		//dc.FillSolidRect(rect,MOF_BLUE_COLOR);
		//dc.Rectangle(rect);
		if(m_structFields.GetCount() > 0) //special stuff for struct
		{
			CPoint pt(m_sx+MOF_DECORATOR_MARGINX,m_sy+MOF_DECORATOR_MARGINY);
			d_util.DrawText(&dc,m_name,pt,d_util.GetFont(GME_PORTNAME_FONT),m_nameColor, TA_TOP | TA_LEFT);
			CSize ext = dc.GetTextExtent(m_name);
			POSITION pos;
			for( pos = m_structFields.GetHeadPosition(); pos != NULL; )
			{
				CString line = m_structFields.GetNext(pos);
				pt.y += ext.cy;
				d_util.DrawText(&dc,line,pt,d_util.GetFont(GME_PORTNAME_FONT),m_nameColor, TA_TOP | TA_LEFT);
			}
			dc.Detach();
			return S_OK;
		}
	}

	if(m_isClass || m_isAspect)
	{
		//d_util.DrawFlatBox(&dc, rect, m_color, MOF_TRANSPARENT_COLOR);
		dc.SelectObject(d_util.GetPen(m_isActive ? m_color : GME_GRAYED_OUT_COLOR));
		dc.SelectStockObject(NULL_BRUSH);
		dc.Rectangle(m_sx, m_sy, m_ex, m_ey);

		CPoint pt((m_sx+m_ex)/2,m_sy+MOF_DECORATOR_MARGINY);
		d_util.DrawText(&dc,m_name,pt,d_util.GetFont(m_isAbstract ? GME_ABSTRACT_FONT: GME_PORTNAME_FONT),m_nameColor, TA_TOP | TA_CENTER);
		pt.x = m_sx+MOF_DECORATOR_MARGINX;
		CSize ext = dc.GetTextExtent(m_name);

		//draw line
		dc.MoveTo(m_sx, m_sy + MOF_DECORATOR_MARGINY + ext.cy);
		dc.LineTo(m_ex, m_sy + MOF_DECORATOR_MARGINY + ext.cy);

		POSITION pos;
		for( pos = m_classAttributes.GetHeadPosition(); pos != NULL; )
		{
			CString line = m_classAttributes.GetNext(pos);
			pt.y += ext.cy;
			d_util.DrawText(&dc,line,pt,d_util.GetFont(GME_PORTNAME_FONT),m_nameColor, TA_TOP | TA_LEFT);
		}
		//draw line again
		dc.MoveTo(m_sx, pt.y+ext.cy);
		dc.LineTo(m_ex, pt.y+ext.cy);
		for( pos = m_classOperations.GetHeadPosition(); pos != NULL; )
		{
			CString line = m_classOperations.GetNext(pos);
			pt.y += ext.cy;
			d_util.DrawText(&dc,line,pt,d_util.GetFont(GME_PORTNAME_FONT),m_nameColor, TA_TOP | TA_LEFT);
		}
		//dc.Detach();
		//return S_OK;  //fall through to proxy


	}

	if(m_isAttr)
	{
		d_util.DrawFlatBox(&dc, rect, m_color, MOF_GREEN_COLOR);
		//dc.FillSolidRect(rect,MOF_GREEN_COLOR);
		//dc.Rectangle(rect);
	}

	if(m_isParameter || m_isOperation)
	{
		d_util.DrawFlatBox(&dc, rect, m_color, MOF_GREY_COLOR);
		//dc.FillSolidRect(rect,MOF_GREY_COLOR);
		//dc.Rectangle(rect);
	}

	if(m_isPackage)
	{
		if(m_ex-m_sx > MOF_PACKAGE_TOP_WIDTH && m_ey-m_sy > MOF_PACKAGE_TOP_HEIGHT)
		{	
			d_util.DrawFlatBox(&dc, CRect(m_sx,m_sy,m_sx+MOF_PACKAGE_TOP_WIDTH,m_sy+MOF_PACKAGE_TOP_HEIGHT), m_color, MOF_TRANSPARENT_COLOR);
			d_util.DrawFlatBox(&dc, CRect(m_sx,m_sy+MOF_PACKAGE_TOP_HEIGHT,m_ex,m_ey), m_color, MOF_TRANSPARENT_COLOR);
			//dc.Rectangle(m_sx,m_sy,m_sx+MOF_PACKAGE_TOP_WIDTH,m_sy+MOF_PACKAGE_TOP_HEIGHT);
			//dc.Rectangle(m_sx,m_sy+MOF_PACKAGE_TOP_HEIGHT,m_ex,m_ey);
		}
	}

	if(m_isConst)
	{
		//d_util.DrawFlatBox(&dc, rect, m_color, MOF_TRANSPARENT_COLOR);
		dc.SelectObject(d_util.GetPen(m_isActive ? m_color : GME_GRAYED_OUT_COLOR));
		dc.SelectStockObject(NULL_BRUSH);
		dc.Rectangle(m_sx, m_sy, m_ex, m_ey);
		
		CPoint pt(m_sx+MOF_DECORATOR_MARGINX,(m_sy+m_ey)/2);
		d_util.DrawText(&dc, CString("const"),pt,d_util.GetFont(GME_PORTNAME_FONT), RGB(0x00,0x00,0xFF), TA_BOTTOM | TA_LEFT);
		CSize ext = dc.GetTextExtent("const ");
		pt.x += ext.cx;
		d_util.DrawText(&dc, m_name,pt,d_util.GetFont(GME_PORTNAME_FONT), m_nameColor, TA_BOTTOM | TA_LEFT);
		dc.Detach();
		return S_OK;
	}

	// Draw Copy Sign (or Import sign)
	if (m_isCopy && m_copyBitmap.IsValid()) {
		CPoint cpt = m_copySignPos;
		cpt.x = m_sx + scalex * m_copySignPos.x;
		cpt.y = m_sy + scaley * m_copySignPos.y;
		m_copyBitmap.DrawTransparent(&dc, cpt.x, cpt.y, MOF_TRANSPARENT_COLOR, !m_isActive, GME_GRAYED_OUT_COLOR);
	}

	if(!m_isException && !m_isClass && !m_isAspect)
		d_util.DrawText(&dc, m_name, rect.CenterPoint(), d_util.GetFont(m_isAbstract ? GME_ABSTRACT_FONT: GME_PORTNAME_FONT), m_nameColor, TA_BOTTOM | TA_CENTER);
	dc.Detach();
	return S_OK;
}

STDMETHODIMP CDecorator::SaveState()
{
	//
	// TODO: The only method where we are in read-write transaction. Store all permanent information
	// (currently GME does not support this) 
	//
	VERIFY_INIT;
	return S_OK;
}


//////////// Decorator private functions
CDecorator::CDecorator() : m_sx(0), m_sy(0), m_ex(0), m_ey(0), 
	m_isActive(true), m_mgaFco(0), m_metaFco(0),
	m_isInitialized(false), m_isLocSet(false), m_isInheritance(false), m_isAssociation(false), 
	m_isConstraint(false), m_isCopy(false), m_isType(false), m_isAttr(false),
	m_isOperation(false), m_isParameter(false), m_isStructField(false), m_isTag(false),
	m_isPackage(false), m_isConst(false), m_isImported(false), m_isClass(false),
	m_isException(false), m_isAbstract(false), m_isAspect(false), m_isSameAspect(false),
	m_calcSize(DWORD(0)), m_copySignPos(DWORD(0)),
	m_color(GME_BLACK_COLOR), m_nameColor(GME_BLACK_COLOR)
{
}

CDecorator::~CDecorator()
{
}

bool CDecorator::GetMetaFCO(const CComPtr<IMgaMetaPart> &metaPart, CComPtr<IMgaMetaFCO> &metaFco)
{
	if (!metaPart) {
		return false;
	}

	metaFco = NULL;
	CComPtr<IMgaMetaRole> metaRole;
	try {
		COMTHROW(metaPart->get_Role(&metaRole));
		COMTHROW(metaRole->get_Kind(&metaFco));
	}
	catch (hresult_exception &) {
		metaFco = NULL;
	}
	return (metaFco != NULL);
}

bool CDecorator::GetPreference(CString &val,const CString &path)
{
	CComBSTR pathBstr(path);
	CComBSTR bstrVal;
	if (m_mgaFco) {
		COMTHROW(m_mgaFco->get_RegistryValue(pathBstr,&bstrVal));
	}
	else {
		COMTHROW(m_metaFco->get_RegistryValue(pathBstr,&bstrVal));
	}
	val = bstrVal;
	return !val.IsEmpty();
}

bool CDecorator::GetPreference(int &val,const CString &path,bool hex)
{
	CString strVal;
	GetPreference(strVal, path);
	return (sscanf(strVal,hex ? "%x" : "%d",&val) == 1);
}

bool CDecorator::GetColorPreference(unsigned long &color, const CString &path)
{
	int i;
	if(GetPreference(i,path,true)) {
		unsigned int r = (i & 0xff0000) >> 16;
		unsigned int g = (i & 0xff00) >> 8;
		unsigned int b = i & 0xff;
		color = RGB(r,g,b);
		return true;
	}
	return false;
}

bool CDecorator::SetupInheritance()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try {
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_INHERITANCE_NAME) {
			m_isInheritance = true;
			m_calcSize.cx = MOF_INHERITANCE_WIDTH;
			m_calcSize.cy = MOF_INHERITANCE_HEIGHT;
			return true;
		}
	}
	catch (hresult_exception &) {
	}
	return false;
}

bool CDecorator::SetupAssociation()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try {
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_ASSOCIATION_NAME) {
			m_isAssociation = true;
			m_calcSize.cx = MOF_ASSOCIATION_WIDTH;
			m_calcSize.cy = MOF_ASSOCIATION_HEIGHT;
			return true;
		}
	}
	catch (hresult_exception &) {
	}
	return false;
}

bool CDecorator::SetupConstraint()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try {
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_CONSTRAINT_NAME) {
			m_isConstraint = true;
			m_constraintBitmap.ReadFromResource(IDB_BITMAP_CONSTRAINT);
			if (m_constraintBitmap.IsValid()) {
				m_calcSize.cx = m_constraintBitmap.Width();
				m_calcSize.cy = m_constraintBitmap.Height();
				return true;
			}
		}
	}
	catch (hresult_exception &) {
	}
	return false;
}

bool CDecorator::SetupType()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try 
	{
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_ALIASTYPE_NAME)
		{
			m_isType = true;
			if(m_isCopy)
			{
				CString proxyName;
				if(GetProxyDisplayName(proxyName))
				{
					m_name = proxyName;
				}
			}
			//setup special stuff for the name
			return true;
		}
		if (name == MOF_COLLECTIONTYPE_NAME)
		{
			m_isType = true;
			if(m_isCopy)
			{
				CString proxyName;
				if(GetProxyDisplayName(proxyName))
				{
					m_name = proxyName;
				}
			}
			//setup special stuff for the name
			CString type;
			if(GetTypeName(type,m_mgaFco))
				m_name = m_name + " : {" + type + "}";
			return true;
		}
		if (name == MOF_ENUMTYPE_NAME)
		{
			m_isType = true;
			if(m_isCopy)
			{
				CString proxyName;
				if(GetProxyDisplayName(proxyName))
				{
					m_name = proxyName;
				}
			}
			//setup special stuff for the name
			if(m_mgaFco)
				m_name = "enum "+m_name;
			return true;
		}
		if (name == MOF_PRIMITIVETYPE_NAME)
		{
			m_isType = true;
			if(m_isCopy)
			{
				CString proxyName;
				if(GetProxyDisplayName(proxyName))
				{
					m_name = proxyName;
				}
			}
			//setup special stuff for the name
			CString type;
			if(GetTypeName(type,m_mgaFco))
				m_name = m_name + " : " + type;
			return true;
		}
		if (name == MOF_STRUCTURETYPE_NAME)
		{
			m_isType = true;
			if(m_isCopy)
			{
				CString proxyName;
				if(GetProxyDisplayName(proxyName))
				{
					m_name = proxyName;
				}
			}
			//setup special stuff for the name
			if(GetTypeStrings())
				m_name = m_name + " {";
			else
				m_name = m_name + " {}";
			return true;
		}
		if (name == MOF_DATATYPEPROXY_NAME)
		{
			m_isType = true;
			//a datatypeproxy that refers to null, no special setup
			return true;
		}
	}
	catch (hresult_exception &) {}
	return false;
}

bool CDecorator::SetupAttrParamSField()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try 
	{
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_ATTRIBUTE_NAME)
		{
			m_isAttr = true;
		}
		if (name == MOF_PARAMETER_NAME)
		{
			m_isParameter = true;
		}
		if (name == MOF_STRUCTFIELD_NAME)
		{
			m_isStructField = true;
		}
		if(m_isAttr || m_isParameter || m_isStructField)
		{
			//common name setup stuff
			if(m_mgaFco)
			{
				if(m_isCopy)
				{
					CString proxyName;
					if(GetProxyDisplayName(proxyName))
					{
						m_name = proxyName;
					}
				}
				CString typeName;
				if(GetTypeName(typeName,m_mgaFco))
					m_name = m_name + " : " + typeName;
			}
			return true;
		}
	}
	catch(hresult_exception &){}
	return false;

}

bool CDecorator::SetupOperation()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try 
	{
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_OPERATION_NAME)
		{
			m_isOperation = true;

			CString opName;
			if(GetOperationString(opName,m_mgaFco))
				m_name = opName;

			return true;
		}
	}
	catch(hresult_exception &){}
	return false;
}

bool CDecorator::SetupTag()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try {
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_TAG_NAME) {
			m_isTag = true;
			m_tagBitmap.ReadFromResource(IDB_BITMAP_TAG);
			if (m_tagBitmap.IsValid()) {
				m_calcSize.cx = m_tagBitmap.Width();
				m_calcSize.cy = m_tagBitmap.Height();
				return true;
			}
		}
	}
	catch (hresult_exception &) {
	}
	return false;
}

bool CDecorator::SetupImported()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try {
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_IMPORTED_NAME) {
			m_isImported = true;
			m_importedBitmap.ReadFromResource(IDB_BITMAP_IMPORTED);
			if (m_importedBitmap.IsValid()) {
				m_calcSize.cx = m_importedBitmap.Width();
				m_calcSize.cy = m_importedBitmap.Height();
				return true;
			}
		}
	}
	catch (hresult_exception &) {
	}
	return false;
}

bool CDecorator::SetupPackage()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try {
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_PACKAGE_NAME || name == MOF_NESTEDPACKAGE_NAME) {
			m_isPackage = true;
			return true;
		}
	}
	catch (hresult_exception &){}
	return false;

}

bool CDecorator::SetupConst()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try 
	{
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_CONSTANT_NAME)
		{
			m_isConst = true;
			return true;
		}
	}
	catch(hresult_exception &){}
	return false;

}

bool CDecorator::SetupClass()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try 
	{
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_CLASS_NAME || name == MOF_CLASSPROXY_NAME)
		{
			m_isClass = true;
			//special name stuff
			GetTypeStrings();
			if(m_isCopy)
			{
				CString proxyName;
				if(GetProxyDisplayName(proxyName))
				{
					m_name = proxyName;
				}
			}
			return true;
		}
	}
	catch(hresult_exception &){}
	return false;
}

bool CDecorator::SetupException()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try 
	{
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_EXCEPTION_NAME || name == MOF_EXCEPTIONPROXY_NAME) 
		{
			m_isException = true;
			if(m_isCopy)
			{
				CString proxyName;
				if(GetProxyDisplayName(proxyName))
				{
					m_name = proxyName;
				}
			}
			return true;

		}
	}
	catch (hresult_exception &) {
	}
	return false;
}

bool CDecorator::SetupAspect()
{
	ASSERT(m_metaFco);
	CComBSTR	bstr;
	try 
	{
		COMTHROW(m_metaFco->get_Name(&bstr));
		CString name(bstr);
		if (name == MOF_ASPECT_NAME || name == MOF_ASPECTPROXY_NAME) 
		{
			m_isAspect = true;
			return true;

		}
		if (name == MOF_SAMEASPECT_NAME)
		{
			m_isSameAspect = true;
			m_calcSize.cx = MOF_SAMEASPECT_WIDTH;
			m_calcSize.cy = MOF_SAMEASPECT_HEIGHT;
			return true;
		}
	}
	catch (hresult_exception &) {
	}
	return false;
}

void CDecorator::CalcRelPositions(CDC *pDC)
{
	if(m_isImported && !m_mgaFco) //import that points to nothing..
		return;
	int maxHeight = 0, maxWidth = 0;
	CSize ext;
	CDC	dc;
	
	dc.Attach(pDC ? pDC->m_hDC : GetDC(NULL));			// Trick
	CFont *oldfont = dc.SelectObject(d_util.GetFont(GME_PORTNAME_FONT));
	
	ext = dc.GetTextExtent(m_name);
	maxWidth = max(maxWidth, ext.cx);
	maxHeight = max(maxHeight, ext.cy);

	if(m_isConst)
	{
		ext = dc.GetTextExtent("const ");
		maxWidth += ext.cx;
	}

	if(m_isException)
	{
		ext.cy = ext.cy + MOF_EXCEPTION_HEIGHT;
		maxHeight = ext.cy;
		if (m_isCopy && m_copyBitmap.IsValid())
		{
			ext.cx = ext.cx + 2* m_copyBitmap.Width();
			maxWidth = max(maxWidth, ext.cx);
		}
	}

	//make the box bigger for objects with small or null names
	ext = dc.GetTextExtent("MINSIZE");
	maxWidth = max(maxWidth, ext.cx);
	maxHeight = max(maxHeight, 3*ext.cy);

	
	if(m_isType && m_structFields.GetCount() > 0) //a struct, 
	{
		int lineCount = 1; //start at one to add in the name line
		POSITION pos;
		for( pos = m_structFields.GetHeadPosition(); pos != NULL; )
		{
			CString line = m_structFields.GetNext( pos );
			ext = dc.GetTextExtent(line);
			maxWidth = max(maxWidth, ext.cx);
			lineCount++;
		}

		maxHeight = max(maxHeight, ext.cy * lineCount);

	}

	if(m_isClass)
	{
		int lineCount = 1; //start at one to add in the name line
		if(m_isCopy) lineCount = 2; //add in a line to give room for the copy/import icon
		POSITION pos;
		for( pos = m_classAttributes.GetHeadPosition(); pos != NULL; )
		{
			CString line = m_classAttributes.GetNext( pos );
			ext = dc.GetTextExtent(line);
			maxWidth = max(maxWidth, ext.cx);
			lineCount++;
		}

		for( pos = m_classOperations.GetHeadPosition(); pos != NULL; )
		{
			CString line = m_classOperations.GetNext( pos );
			ext = dc.GetTextExtent(line);
			maxWidth = max(maxWidth, ext.cx);
			lineCount++;
		}

		maxHeight = max(maxHeight, ext.cy * lineCount);
	}


	m_calcSize.cx = maxWidth + 2 * MOF_DECORATOR_MARGINX;
	m_calcSize.cy = (maxHeight + 2 * MOF_DECORATOR_MARGINY);

	if(m_isPackage)
	{
		m_calcSize.cy = MOF_PACKAGE_HEIGHT;
		m_calcSize.cx = max(m_calcSize.cx,MOF_PACKAGE_TOP_WIDTH*2);
	}

	int xcenterpos = (2 * MOF_DECORATOR_MARGINX + maxWidth) / 2; 
	int	xleftpos = MOF_DECORATOR_MARGINX;
	int	xrightpos = MOF_DECORATOR_MARGINX + maxWidth;
	int ypos = MOF_DECORATOR_MARGINY;

	if (m_isCopy && m_copyBitmap.IsValid()) {
		ypos = m_calcSize.cy;
		ypos -= MOF_DECORATOR_MARGINY;
		ypos -= m_copyBitmap.Height();
		m_copySignPos.x = xleftpos;
		m_copySignPos.y = ypos;
		
	}

	if (pDC) {
		dc.SelectObject(oldfont);
		dc.Detach();
	}
	//more stuff to come..
}

bool CDecorator::GetAttribute(CString &val, const CString &attrname, CComPtr<IMgaFCO> mgaFco)
{
	if (!mgaFco) {
		mgaFco = m_mgaFco;
	}
	if (!mgaFco) {
		return false;
	}
	CComBSTR attr(attrname);
	CComBSTR bstrVal;
	try {
		COMTHROW(mgaFco->get_StrAttrByName(attr,&bstrVal));
	}
	catch (hresult_exception &) {
		bstrVal.Empty();
	}
	val = bstrVal;
	return !val.IsEmpty();
}

bool CDecorator::GetAttribute(bool &val, const CString &attrname, CComPtr<IMgaFCO> mgaFco)
{
	if (!mgaFco) {
		mgaFco = m_mgaFco;
	}
	if (!mgaFco) {
		return false;
	}
	CComBSTR attr(attrname);
	VARIANT_BOOL vval;
	try {
		COMTHROW(mgaFco->get_BoolAttrByName(attr,&vval));
	}
	catch (hresult_exception &) {
		return false;
	}
	val = (vval == VARIANT_TRUE);
	return true;
}

CComPtr<IMgaFCO> CDecorator::GetTypeFCO(CComPtr<IMgaFCO> mgaFco)
{
	if(!mgaFco)
		return NULL;

	CComPtr<IMgaConnPoints>	connPoints;
	COMTHROW(mgaFco->get_PartOfConns(&connPoints));
	CComPtr<IMgaConnPoint> connPoint;
	MGACOLL_ITERATE(IMgaConnPoint,connPoints) 
	{
		connPoint = MGACOLL_ITER;

		CComBSTR bstrRole;
		COMTHROW(connPoint->get_ConnRole(&bstrRole));

		CString role(bstrRole);

		if( role == "dst")
		{
			CComPtr<IMgaConnection> conn;	
			COMTHROW(connPoint->get_Owner(&conn));
			
			CComPtr<IMgaMetaFCO> metaFco;
			COMTHROW(conn->get_Meta(&metaFco));
			CComBSTR bstrMetaName;
			COMTHROW(metaFco->get_Name(&bstrMetaName));

			CString metaName(bstrMetaName);

			if( metaName == "IsOfType")
			{
				CComPtr<IMgaConnPoints>	targetConnPoints;

				COMTHROW(conn->get_ConnPoints(&targetConnPoints));
				CComPtr<IMgaConnPoint> targetConnPoint;
				MGACOLL_ITERATE(IMgaConnPoint,targetConnPoints) 
				{
					targetConnPoint = MGACOLL_ITER;
					if (!connPoint.IsEqualObject(targetConnPoint)) 
					{
						CComPtr<IMgaFCO> targetFco;
						COMTHROW(targetConnPoint->get_Target(&targetFco));
						CComPtr<IMgaReference> ref;
						targetFco.QueryInterface(&ref);
						if(ref)
						{
							CComPtr<IMgaFCO> referred;
							COMTHROW(ref->get_Referred(&referred));
							return referred;
						}
						return targetFco;
					}

				} MGACOLL_ITERATE_END;
			}
		}

	} MGACOLL_ITERATE_END;

	return NULL; 
}

bool CDecorator::GetTypeName(CString &val, CComPtr<IMgaFCO> mgaFco)
{
	if(mgaFco)
	{
		try
		{
			CComPtr<IMgaFCO> type =  GetTypeFCO(mgaFco);
			if(type)
			{
				CComBSTR bstr;
				CComPtr<IMgaMetaFCO> meta;
				COMTHROW(type->get_Meta(&meta));
				COMTHROW(meta->get_Name(&bstr));
				CString metaName(bstr);
				if(metaName != "PrimitiveType")
				{
					COMTHROW(type->get_Name(&bstr));
					CString name(bstr);
					val = name;
					return true;
				}
				else //followed connection to primitive type
				{
					if(GetAttribute(val, "type",type))
						return true;
				}
			}
			else//mgaFco itself is a primitive type
			{
				if(GetAttribute(val,"type",mgaFco))
					return true;
			}
		}
		catch(hresult_exception &){}
	}
	

	return false;

}

bool CDecorator::GetTypeStrings()
{
	bool ret = false;
	if(!m_mgaFco)
		return ret;

	CComPtr<IMgaModel> mgaModel;
	m_mgaFco.QueryInterface(&mgaModel);
	if(!mgaModel)
		return ret;

	CComPtr<IMgaFCOs> children;
	COMTHROW(mgaModel->get_ChildFCOs(&children));

	MGACOLL_ITERATE(IMgaFCO, children)
	{
		CComPtr<IMgaFCO> child = MGACOLL_ITER;
		CComPtr<IMgaMetaFCO> metaFco;
		COMTHROW(child->get_Meta(&metaFco));
		CComBSTR bstrMetaName;
		COMTHROW(metaFco->get_Name(&bstrMetaName));

		CString metaName(bstrMetaName);

		if( (metaName == MOF_ATTRIBUTE_NAME) || (metaName == MOF_OPERATION_NAME) || (metaName == MOF_STRUCTFIELD_NAME) )
		{
				
			CComBSTR bstr;
			COMTHROW(child->get_Name(&bstr));
			CString name(bstr);
			CString type;
			if(GetTypeName(type,child))
				name = name + " : " + type;
			if(m_isType)
				m_structFields.AddTail(name);
			if(m_isClass)
			{
				CString op;
				if(GetOperationString(op,child))
					m_classOperations.AddTail(op);
				else
					m_classAttributes.AddTail(name);
			}
			ret = true;
		}

	} MGACOLL_ITERATE_END;

	if(ret)
		m_structFields.AddTail("}");
	return ret; 
}

bool CDecorator::GetOperationString(CString &val, CComPtr<IMgaFCO> mgaFco)
{
	bool ret = false;
	if(!mgaFco)
		return ret;

	CComPtr<IMgaMetaFCO> pmetaFco;
	COMTHROW(mgaFco->get_Meta(&pmetaFco));

	CComBSTR bstrName;

	COMTHROW(pmetaFco->get_Name(&bstrName));
	CString name(bstrName);
	if (name != MOF_OPERATION_NAME)
		return ret;
	ret = true;

	COMTHROW(mgaFco->get_Name(&bstrName));
	val = bstrName;
	val = val + "(";

	CString returnType = "void";
	bool first = true;

	CComPtr<IMgaFCOs> children;
	CComPtr<IMgaModel> mgaModel;
	mgaFco.QueryInterface(&mgaModel);
	if(!mgaModel)
		return false;
	COMTHROW(mgaModel->get_ChildFCOs(&children));

	MGACOLL_ITERATE(IMgaFCO, children)
	{

		CComPtr<IMgaFCO> child = MGACOLL_ITER;
		CComPtr<IMgaMetaFCO> childMeta;
		COMTHROW(child->get_Meta(&childMeta));
		CComBSTR bstrChildMetaName;
		COMTHROW(childMeta->get_Name(&bstrChildMetaName));
		CString childMetaName(bstrChildMetaName);
		if(childMetaName == MOF_PARAMETER_NAME)
		{
			CComBSTR bstr;
			COMTHROW(child->get_Name(&bstr));
			CString name(bstr);
			CString type;
			if(GetTypeName(type,child))
				name = name + " : " + type;
			
			CString direction = "";
			if(GetAttribute(direction,"direction",child) && (direction == "return"))
				returnType = type;
			else
			{
				if(!first)
					val = val + ",";
				else
					first = false;
				val = val + " " + name;
			}

		}

	} MGACOLL_ITERATE_END;

	val = val + ") : " + returnType;

	return ret;
}

bool CDecorator::GetProxyDisplayName(CString &val)
{

	if(!m_mgaFco)
		return false;

	CComBSTR bstrName;
	COMTHROW(m_mgaFco->get_Name(&bstrName));
	CString name(bstrName);
	val = name;

	CComPtr<IMgaModel> parent;
	CComPtr<IMgaMetaFCO> parentMeta;
	COMTHROW(m_mgaFco->get_ParentModel(&parent));
	if(!parent)
	{
		return true;
	}

	CComBSTR bstrMetaName;

	COMTHROW(parent->get_Meta(&parentMeta));
	COMTHROW(parentMeta->get_Name(&bstrMetaName));
	CString metaName(bstrMetaName);

	while(metaName == MOF_NESTEDPACKAGE_NAME || metaName == MOF_CLASS_NAME)
	{
		COMTHROW(parent->get_Name(&bstrName));
		name = CString(bstrName);
		val = name + " :: " + val;
		
		CComPtr<IMgaModel> grandparent;
		COMTHROW(parent->get_ParentModel(&grandparent));
		if(!grandparent)
			return true;
		parent = grandparent;

		CComPtr<IMgaMetaFCO> grandparentMeta; //for some strange reason, couldn't use parentMeta or bstrMetaName from above, was throwing com exceptions.. but declaring these new variables works.... 
		COMTHROW(parent->get_Meta(&grandparentMeta));

		CComBSTR grandparentMetaName;
		COMTHROW(grandparentMeta->get_Name(&grandparentMetaName));
		metaName = CString(grandparentMetaName);
	}

	if(metaName == MOF_PACKAGE_NAME) //reached the top
	{
		COMTHROW(parent->get_Name(&bstrName));
		name = CString(bstrName);
		val = name + " :: " + val;
	}
	return true;

	
}