// Decorator.h : Declaration of the CDecorator

#ifndef __MOFDECORATOR_H_
#define __MOFDECORATOR_H_

#include "DecoratorStd.h"
#include "DecoratorConfig.h"
#include "Resource.h"       // main symbols
#include "DecoratorLib.h"
#include "DecoratorUtil.h"
#include "MaskedBitmap.h"


#define MOF_INHERITANCE_NAME		"Inheritance"
#define MOF_ASSOCIATION_NAME		"Association"
#define	MOF_CONSTRAINT_NAME			"Constraint"
#define MOF_ALIASTYPE_NAME			"AliasType"
#define MOF_COLLECTIONTYPE_NAME		"CollectionType"
#define MOF_ENUMTYPE_NAME			"EnumerationType"
#define MOF_PRIMITIVETYPE_NAME		"PrimitiveType"
#define MOF_STRUCTURETYPE_NAME		"StructureType"
#define MOF_DATATYPEPROXY_NAME		"DataTypeProxy"
#define MOF_ATTRIBUTE_NAME			"Attribute"
#define MOF_PARAMETER_NAME			"Parameter"
#define MOF_STRUCTFIELD_NAME		"StructureField"
#define MOF_OPERATION_NAME			"Operation"
#define MOF_TAG_NAME				"Tag"
#define MOF_PACKAGE_NAME			"Package"
#define MOF_NESTEDPACKAGE_NAME		"NestedPackage"
#define MOF_CONSTANT_NAME			"Constant"
#define MOF_IMPORTED_NAME			"Import"
#define MOF_CLASS_NAME				"Class"
#define MOF_CLASSPROXY_NAME			"ClassProxy"
#define MOF_EXCEPTION_NAME			"Exception"
#define MOF_EXCEPTIONPROXY_NAME		"ExceptionProxy"
#define MOF_ASPECT_NAME				"Aspect"
#define MOF_ASPECTPROXY_NAME		"AspectProxy"
#define MOF_SAMEASPECT_NAME			"SameAspect"

#define MOF_ABSTRACT_ATTR			"isAbstract"

#define  MOF_INHERITANCE_WIDTH		24
#define	 MOF_INHERITANCE_HEIGHT		0.866 * MOF_INHERITANCE_WIDTH
#define  MOF_ASSOCIATION_WIDTH		6
#define  MOF_ASSOCIATION_HEIGHT		MOF_ASSOCIATION_WIDTH
#define  MOF_PACKAGE_HEIGHT			70
#define  MOF_PACKAGE_TOP_HEIGHT		10
#define  MOF_PACKAGE_TOP_WIDTH		3 * MOF_PACKAGE_TOP_HEIGHT
#define  MOF_EXCEPTION_HEIGHT		40
#define  MOF_EXCEPTION_RADIUS		4
#define  MOF_SAMEASPECT_HEIGHT		22
#define  MOF_SAMEASPECT_WIDTH		2 * MOF_SAMEASPECT_HEIGHT

#define MOF_GREY_COLOR				RGB(0xDD, 0xDD, 0xDD)
#define MOF_GREEN_COLOR				RGB(0xDD, 0xFF, 0xDD)
#define MOF_BLUE_COLOR				RGB(0xDD, 0xDD, 0xFF)

#define MOF_TRANSPARENT_COLOR		RGB(0xFF, 0xFF, 0xFF)

#define	 MOF_DECORATOR_MARGINX		6
#define	 MOF_DECORATOR_MARGINY		4

/////////////////////////////////////////////////////////////////////////////
// CDecorator
class ATL_NO_VTABLE CDecorator : 
	public CComObjectRootEx<CComSingleThreadModel>,
	public IMgaDecorator,
	public CComCoClass<CDecorator, &CLSID_Decorator>
{
public:
	CDecorator();
	~CDecorator();

DECLARE_REGISTRY_RESOURCEID(IDR_DECORATOR)

DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(CDecorator)
	COM_INTERFACE_ENTRY(IMgaDecorator)
END_COM_MAP()

// IMgaDecorator
public:
	STDMETHOD(Initialize)(/*[in]*/ IMgaProject *project, /*[in]*/ IMgaMetaPart *meta, /*[in]*/ IMgaFCO *obj);
	STDMETHOD(Destroy)();
	STDMETHOD(GetMnemonic)(/*[out]*/ BSTR *mnemonic);
	STDMETHOD(GetFeatures)(/*[out]*/ feature_code *features);
	STDMETHOD(SetParam)(/*[in]*/ BSTR name, /*[in]*/ VARIANT value);
	STDMETHOD(GetParam)(/*[in]*/ BSTR name, /*[out]*/ VARIANT* value);
	STDMETHOD(SetActive)(/*[in]*/ VARIANT_BOOL isActive);
	STDMETHOD(GetPreferredSize)(/*[out]*/ long* sizex, /*[out]*/ long* sizey);
	STDMETHOD(SetLocation)(/*[in]*/ long sx, /*[in]*/ long sy, /*[in]*/ long ex, /*[in]*/ long ey);
	STDMETHOD(GetLocation)(/*[out]*/ long *sx, /*[out]*/ long *sy, /*[out]*/ long *ex, /*[out]*/ long *ey);	
	STDMETHOD(GetLabelLocation)( /*[out]*/ long *sx, /*[out]*/ long *sy, /*[out]*/ long *ex, /*[out]*/ long *ey);
	STDMETHOD(GetPortLocation)(/*[in]*/ IMgaFCO *fco, /*[out]*/ long *sx, /*[out]*/ long *sy, /*[out]*/ long *ex, /*[out]*/ long *ey);
	STDMETHOD(GetPorts)(/*[out, retval]*/ IMgaFCOs **portFCOs);
	STDMETHOD(Draw)(/*[in]*/ HDC hdc);
	STDMETHOD(SaveState)();
	
protected:
	bool m_isInitialized;
	bool m_isLocSet;
	long m_sy;
	long m_sx;
	long m_ey;
	long m_ex;
	bool m_isActive;
	bool m_isCopy;
	bool m_isInheritance;
	bool m_isAssociation;
	bool m_isConstraint;
	bool m_isType;
	bool m_isAttr;
	bool m_isOperation;
	bool m_isParameter;
	bool m_isStructField;
	bool m_isTag;
	bool m_isPackage;
	bool m_isConst;
	bool m_isImported;
	bool m_isClass;
	bool m_isException;
	bool m_isAbstract;
	bool m_isAspect;
	bool m_isSameAspect;
	CComPtr<IMgaFCO>	m_mgaFco;
	CComPtr<IMgaMetaFCO>	m_metaFco;
	CString		m_name;
	COLORREF	m_color;
	COLORREF	m_nameColor;
	CSize	m_calcSize;
	CPoint	m_copySignPos;
	CMaskedBitmap m_constraintBitmap;
	CMaskedBitmap m_tagBitmap;
	CMaskedBitmap m_copyBitmap;
	CMaskedBitmap m_importedBitmap;
	CStringList m_classAttributes;
	CStringList m_classOperations;
	CStringList m_structFields;

private:
	bool GetMetaFCO(const CComPtr<IMgaMetaPart> &metaPart, CComPtr<IMgaMetaFCO> &metaFco);
	bool GetPreference(CString &val, const CString &path);
	bool GetPreference(int &val, const CString &path, bool hex = false);
	bool GetColorPreference(COLORREF &color, const CString &path);
	bool SetupInheritance();
	bool SetupAssociation();
	bool SetupConstraint();
	bool SetupType();
	bool SetupAttrParamSField();
	bool SetupOperation();
	bool SetupTag();
	bool SetupPackage();
	bool SetupConst();
	bool SetupImported();
	bool SetupClass();
	bool SetupException();
	bool SetupAspect();

	void CalcRelPositions(CDC *pDC = NULL);
	bool GetAttribute(CString &val, const CString &attrname, CComPtr<IMgaFCO> mgaFco);
	bool GetAttribute(bool &val, const CString &attrname, CComPtr<IMgaFCO> mgaFco);
	CComPtr<IMgaFCO> GetTypeFCO(CComPtr<IMgaFCO> mgaFco);
	bool GetTypeName(CString &val, CComPtr<IMgaFCO> mgaFco);
	CString GetDecoratedString(CComPtr<IMgaFCO> mgaFco);
	bool GetTypeStrings();
	bool GetOperationString(CString &val, CComPtr<IMgaFCO> mgaFco);
	bool GetProxyDisplayName(CString &val);
};


#endif //__MOFDECORATOR_H_
