#include "GetType.h"

std::string GetType(const ExtendedMOF::Classifier & classifier)
{
	std::string type;

	if(classifier.type() == ExtendedMOF::PrimitiveType::meta)
		type = GetType(ExtendedMOF::PrimitiveType::Cast(classifier));
	else if(classifier.type() == ExtendedMOF::CollectionType::meta)
		type = GetType(ExtendedMOF::CollectionType::Cast(classifier));
	else if(classifier.type() == ExtendedMOF::DataTypeProxy::meta)
		type = GetType(ExtendedMOF::DataTypeProxy::Cast(classifier));
	else if(classifier.type() == ExtendedMOF::ClassProxy::meta)
		type = GetType(ExtendedMOF::ClassProxy::Cast(classifier));
	else
		type = classifier.name();

	return type;
}

std::string GetType(const ExtendedMOF::Parameter & parameter)
{
	ExtendedMOF::Classifier classifier(ExtendedMOF::IsOfType(parameter.srcIsOfType()).srcIsOfType_end());
	std::string type(GetType(classifier));	
	std::string direction(parameter.direction());
	std::string cppType;
	
	if(direction == "in")
		cppType = "const " + type + " & ";
	else if(direction == "return")
		cppType = type;
	else
		cppType = type + " & ";

	return cppType;
}

std::string GetType(const ExtendedMOF::StructureField & field)
{
	ExtendedMOF::Classifier classifier(ExtendedMOF::IsOfType(field.srcIsOfType()).srcIsOfType_end());
	return GetType(classifier);
}

std::string GetType(const ExtendedMOF::PrimitiveType & primitive)
{
	std::string type(primitive.type());
	std::string cppType;

	if(type == "Boolean")
		cppType = "bool";
	else if(type == "Integer")
		cppType = "int";
	else if(type == "Double")
		cppType = "double";
	else if(type == "Long")
		cppType = "long";
	else if(type == "Float")
		cppType = "float";
	else
		cppType = "std::string";

	return cppType;
}

std::string GetType(const ExtendedMOF::CollectionType & collection)
{
	ExtendedMOF::Classifier classifier(ExtendedMOF::IsOfType(collection.srcIsOfType()).srcIsOfType_end());
	std::string collectedType(GetType(classifier));

	bool isUnique(collection.isUniqueCT());
	std::string cppType;

	if(isUnique)
		cppType = "std::set<" + collectedType + " &>";
	else
		cppType = "std::multiset<" + collectedType + " &>";

	return cppType;
}

std::string GetType(const ExtendedMOF::DataTypeProxy & proxy)
{
	return GetType(proxy.ref());
}

std::string GetType(const ExtendedMOF::ClassProxy & proxy)
{
	return GetType(proxy.ref());
}